##################################################################################
#
# model.py
# 
# Procedurally generate a U-Net of the specified depth, filter size, & with the 
# specified number of repeating convolutions per convolution block.
#
##################################################################################
#
# Luke Sheneman, 2023
# sheneman@uidaho.edu
#
# Institute for Interdisciplinary Data Studies (IIDS)
# University of Idaho
#
##################################################################################
#
# MIT License
#
# Copyright (c) 2023 Luke Sheneman
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################


from keras.layers import Input, concatenate, Dense, Conv2D, MaxPooling2D, UpSampling2D, BatchNormalization, Dropout
from keras.layers import LeakyReLU

#
# Procedurally generated U-Net model
#
def unet(input_img, unet_depth, nfilter, cb_width):

	print("MAX_UNET_DEPTH = %d" %unet_depth)
	print("TOTAL STEPS = %d" %(unet_depth*2-1))

	convs = []
	pool = input_img
	for step in range(unet_depth*2-1):

		# THE ENCODER
		if(step<unet_depth-1):
			conv = Conv2D(nfilter, 3, activation = LeakyReLU(alpha=0.1), padding = 'same')(pool)
			conv = BatchNormalization()(conv)

			for i in range(cb_width-1):
				conv = Conv2D(nfilter, 3, activation = LeakyReLU(alpha=0.1), padding = 'same')(conv)
				conv = BatchNormalization()(conv)

			pool = MaxPooling2D(pool_size=(2, 2))(conv)
			pool = Dropout(0.5)(pool)
			convs.append(conv)
			nfilter = nfilter * 2

		# THE BOTTLENECK
		elif(step==unet_depth-1):
			conv = Conv2D(nfilter, 3, activation = LeakyReLU(alpha=0.1), padding = 'same')(pool)
			conv = BatchNormalization()(conv)

			for i in range(cb_width-1):
				conv = Conv2D(nfilter, 3, activation = LeakyReLU(alpha=0.1), padding = 'same')(conv)
				conv = BatchNormalization()(conv)

			convs.append(conv)
			nfilter = nfilter / 2

		# THE DECODER
		else:
			up = Conv2D(nfilter, 2, activation = LeakyReLU(alpha=0.1), padding = 'same')(UpSampling2D(size = (2,2))(conv))
			up = BatchNormalization()(up)
			up = Dropout(0.5)(up)

			mapping = (unet_depth*2-2)-step
			merge = concatenate([convs[(unet_depth*2-2)-step],up], axis = 3)    # SKIP CONNECTION

			conv = Conv2D(nfilter, 3, activation = LeakyReLU(alpha=0.1), padding = 'same')(merge)
			conv = BatchNormalization()(conv)

			for i in range(cb_width-1):
				conv = Conv2D(nfilter, 3, activation = LeakyReLU(alpha=0.1), padding = 'same')(conv)
				conv = BatchNormalization()(conv)

			nfilter = nfilter / 2

	decoded = Conv2D(1, 1, activation = 'sigmoid')(conv)
	
	return(decoded)

