##################################################################################
#
# train.py
#
# Trainer script for deep and wide U-Net for generative denoising of 
# ultra-sparse-photon images
#
##################################################################################
#
# Luke Sheneman, 2023
# sheneman@uidaho.edu
#
# Institute for Interdisciplinary Data Studies (IIDS)
# University of Idaho
#
##################################################################################
#
# MIT License
#
# Copyright (c) 2023 Luke Sheneman
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os, sys, time, gc
import numpy as np
import pandas as pd
import datetime

from flow import trainGenerator, valGenerator   # from local dir

np.set_printoptions(linewidth=256)

if(len(sys.argv)!=19):
	print("TRAIN ERROR: INCORRECT NUMBER OF ARGUMENTS.")
	print("  SUPPLIED %d, NEEDED 19" %(len(sys.argv)))
	exit(-1)


NAME		 = sys.argv[1]
OUTPUT_ROOT      = sys.argv[2] 
TRAIN_ROOT_PATH	 = sys.argv[3]
TRAIN_DATA_DIR 	 = sys.argv[4]
TRAIN_TRUTH_DIR  = sys.argv[5]
VAL_ROOT_PATH    = sys.argv[6]
VAL_DATA_DIR     = sys.argv[7]
VAL_TRUTH_DIR    = sys.argv[8]
IMG_SIZE	 = int(sys.argv[9])
EPOCHS           = int(sys.argv[10])
BATCH_SIZE       = int(sys.argv[11])
KFOLD       	 = int(sys.argv[12])
MAX_UNET_DEPTH   = int(sys.argv[13])
START_FILTER_NUM = int(sys.argv[14])
CB_WIDTH	 = int(sys.argv[15])
OPTIMIZER        = sys.argv[16]
LEARNING_RATE    = float(sys.argv[17])
LOSS_FUNCTION    = sys.argv[18]




print("***************************************")
print("")
print("HYPERPARAMETERS:")
print("----------------")
print("              NAME -> %s" %NAME)
print("       OUTPUT_ROOT -> %s" %OUTPUT_ROOT)
print("   TRAIN_ROOT_PATH -> %s" %TRAIN_ROOT_PATH)
print("    TRAIN_DATA_DIR -> %s" %TRAIN_DATA_DIR)
print("   TRAIN_TRUTH_DIR -> %s" %TRAIN_TRUTH_DIR)
print("     VAL_ROOT_PATH -> %s" %VAL_ROOT_PATH)
print("      VAL_DATA_DIR -> %s" %VAL_DATA_DIR)
print("     VAL_TRUTH_DIR -> %s" %VAL_TRUTH_DIR)
print("          IMG_SIZE -> %s" %IMG_SIZE)
print("            EPOCHS -> %d" %EPOCHS)
print("        BATCH_SIZE -> %d" %BATCH_SIZE)
print("             KFOLD -> %d" %KFOLD)
print("    MAX_UNET_DEPTH -> %d" %MAX_UNET_DEPTH)
print("  START_FILTER_NUM -> %d" %START_FILTER_NUM)
print("          CB_WIDTH -> %d" %CB_WIDTH)
print("         OPTIMIZER -> %s" %OPTIMIZER)
print("     LEARNING_RATE -> %f" %LEARNING_RATE)
print("     LOSS_FUNCTION -> %s" %LOSS_FUNCTION)
print("")
print("***************************************")


from sklearn.model_selection import KFold
from keras.layers import Input
from keras.models import Model
import tensorflow.keras.optimizers
from tensorflow.keras.utils import plot_model
import tensorflow as tf

# import the model definition from local model.py
from model2 import unet

import cv2


#SEED = int(time.time())
SEED = 42
print("RANDOM_SEED: ", SEED)
tf.random.set_seed(SEED)


OUTPUT_DIR = os.path.join(OUTPUT_ROOT, NAME)
MODEL_OUT  = os.path.join(OUTPUT_DIR, NAME+".h5")

CHKPT_OUT  	= os.path.join(OUTPUT_DIR, "last.h5")
CHKPT_BEST_OUT	= os.path.join(OUTPUT_DIR, "best.h5")



if(not os.path.isdir(OUTPUT_ROOT)):
	os.mkdir(OUTPUT_ROOT)
if(not os.path.isdir(OUTPUT_DIR)):
	os.mkdir(OUTPUT_DIR)

trainpath = os.path.join(TRAIN_ROOT_PATH,TRAIN_DATA_DIR)
if os.path.isdir(trainpath):
	NUM_TRAIN = len(os.listdir(trainpath))
	print("Found %d training images in %s..." %(NUM_TRAIN,trainpath))
else:
	print("ERROR: Could not find training set in ", trainpath)
	exit(-1)

valpath = os.path.join(VAL_ROOT_PATH,VAL_DATA_DIR)
if os.path.isdir(valpath):
	NUM_VAL = len(os.listdir(valpath))
	print("Found %d validationimages in %s..." %(NUM_VAL,valpath))
else:
	print("ERROR: Could not find validation set in ", valpath)
	exit(-1)
	



# get our model architecture definition
input_img = Input(shape=(IMG_SIZE, IMG_SIZE, 1))
architecture = unet(input_img, MAX_UNET_DEPTH, START_FILTER_NUM, CB_WIDTH)


if(OPTIMIZER == 'Adam' or OPTIMIZER == 'adam'):
	print("HERE!  Adam!")
	opt = tensorflow.keras.optimizers.Adam(learning_rate=LEARNING_RATE);
	print("HERE2!")
elif(OPTIMIZER == 'Adadelta' or OPTIMIZER == 'adadelta'):
	opt = tensorflow.keras.optimizers.Adadelta(learning_rate=LEARNING_RATE);
elif(OPTIMIZER == 'SGD' or OPTIMIZER == 'sgd'):
	opt = tensorflow.keras.optimizers.SGD(learning_rate=LEARNING_RATE);
elif(OPTIMIZER == 'Adamax' or OPTIMIZER == 'adamax'):
	opt = tensorflow.keras.optimizers.Adamax(learning_rate=LEARNING_RATE);
else:
	print("ERROR: Unknown Optimizer %s" %OPTIMIZER)
	exit(-1)



model_checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
	filepath=CHKPT_OUT,
	monitor="loss")

model_checkpoint_best_callback = tf.keras.callbacks.ModelCheckpoint(
	filepath=CHKPT_BEST_OUT,
	save_best_only=True,
	monitor="val_auc",
	#monitor="val_mean_squared_error",
	mode="max")



class CustomCallback(tf.keras.callbacks.Callback):
	def on_train_begin(self, logs=None):
		keys = list(logs.keys())
		print("Starting training; got log keys: {}".format(keys))

	def on_train_end(self, logs=None):
		keys = list(logs.keys())
		print("Stop training; got log keys: {}".format(keys))

	def on_epoch_begin(self, epoch, logs=None):
		keys = list(logs.keys())
		print("Start epoch {} of training; got log keys: {}".format(epoch, keys))

	def on_epoch_end(self, epoch, logs=None):
		keys = list(logs.keys())
		print("End epoch {} of training; got log keys: {}".format(epoch, keys))
		for k in keys:
			print(k + ": " + str(logs[k]))



# define Tensorboard callbacks
log_base = os.path.join(OUTPUT_ROOT, NAME, "logs")
if(not os.path.isdir(log_base)):
	os.mkdir(log_base)
log_dir = os.path.join(log_base, datetime.datetime.now().strftime("%Y%m%d-%H%M%S"))
if(not os.path.isdir(log_dir)):
	os.mkdir(log_dir)
tensorboard_callback = tf.keras.callbacks.TensorBoard(log_dir=log_dir, histogram_freq=1)




traingen = trainGenerator(batch_size = BATCH_SIZE,
			  data_path   = TRAIN_ROOT_PATH,      
			  train_dir   = TRAIN_DATA_DIR,   
			  truth_dir   = TRAIN_TRUTH_DIR, 
			  target_size = (IMG_SIZE,IMG_SIZE),
			  seed	      = SEED)


valgen = valGenerator(batch_size = BATCH_SIZE,
			data_path   = VAL_ROOT_PATH,
			val_dir     = VAL_DATA_DIR,
			truth_dir   = VAL_TRUTH_DIR,
			target_size = (IMG_SIZE,IMG_SIZE),
			seed	    = SEED)




# Train and save the model and performance k-fold cross-validation if needed
if(KFOLD == 0):
	# Instantiate the model (as defined above)
	unet = Model(input_img, architecture)
	unet.compile(optimizer=opt, loss = LOSS_FUNCTION, metrics = [tf.keras.metrics.AUC(), tf.keras.metrics.MeanSquaredError()])
	unet.summary()

	// optionally generate a graph image of our U-Net CNN
	#plot_model(unet, to_file='unet.png', show_shapes=True,show_layer_names=True)

	print("Training.  No cross-validation.")

	print("Before train")
	history = unet.fit(traingen, 
			   validation_data = valgen,
			   validation_steps = NUM_VAL/BATCH_SIZE,
			   steps_per_epoch = int(NUM_TRAIN/BATCH_SIZE),  
			   epochs     	   = EPOCHS,
			   callbacks  	   = [tensorboard_callback, CustomCallback(), model_checkpoint_callback, model_checkpoint_best_callback],
			   shuffle    	   = True)
	print("After train")

	# convert the history.history dict to a pandas DataFrame:     
	hist_df = pd.DataFrame(history.history)

	# save history to csv: 
	hist_csv_file = os.path.join(OUTPUT_DIR, "history.csv")
	print("Saving history to %s" %hist_csv_file)
	with open(hist_csv_file, mode='w') as f:
		hist_df.to_csv(f)


elif(KFOLD > 0):
	print("Training.  k-fold cross-validation with k=%d" %KFOLD)
	kf = KFold(n_splits=KFOLD, shuffle=True)
	fold = 0
	for train_index, val_index in kf.split(X_train):

		# Instantiate the model (as defined above)
		unet = Model(input_img, decoded)
		unet.compile(optimizer=opt, loss = LOSS_FUNCTION, metrics = ['accuracy', tf.keras.metrics.AUC(), tf.keras.metrics.Precision(), tf.keras.metrics.Recall()])
		unet.summary()
		
		print("TRAIN_INDEX: ", train_index)
		print("VAL_INDEX: ", val_index)

		X_train_kfold, X_val_kfold = X_train[train_index], X_train[val_index]
		y_train_kfold, y_val_kfold = y_train[train_index], y_train[val_index]

		print("X_train_kfold: ", X_train_kfold.shape)
		print("  X_val_kfold: ", X_val_kfold.shape)
		print("y_train_kfold: ", y_train_kfold.shape)
		print("  y_val_kfold: ", y_val_kfold.shape)

		history = unet.fit(X_train_kfold, y_train_kfold,
				   batch_size = BATCH_SIZE,
				   epochs     = EPOCHS,
				   validation_data = (X_val_kfold, y_val_kfold),
				   callbacks  = [tensorboard_callback, CustomCallback()],
				   shuffle    = True)

		
		# convert the history.history dict to a pandas DataFrame:
		hist_df = pd.DataFrame(history.history)

		# save history to csv:
		hist_csv_filename = "history_" + str(fold) + ".csv"	
		hist_csv_file = os.path.join(OUTPUT_DIR, hist_csv_filename)
		print("Saving history to %s" %hist_csv_file)
		with open(hist_csv_file, mode='w') as f:
			hist_df.to_csv(f)

		print("Clearing TensorFlow session to save memory")
		tf.keras.backend.clear_session()
		del unet
		for i in range(20):
			gc.collect()

		fold + fold + 1
		
else:
	print("ERROR.  KFOLD = %d is out of range." %KFOLD)
	exit(-1)



