##################################################################################
#
# bootstrap.py
#
# Subsample tiff stacks with parallel multiprocessing to create sparse inputs 
# for training and testing.
#
# Useful for creating training data for our model from high-flux samples.
# Takes a folder of sparse photon centroid tiff stacks produced by centroid.py 
# and randomly samples with replacement P% of all images within the stack.   
# These are then integrated with post-processing to reconstruct the target 
# based on the subsample.  This is repeated n times to produce a new tiff 
# stack which should # represent different views of the same target in each 
# frame. This is useful for creating training data for our model.
#
##################################################################################
#
# Luke Sheneman, 2023
# sheneman@uidaho.edu
#
# Institute for Interdisciplinary Data Studies (IIDS)
# University of Idaho
#
##################################################################################
#
# MIT License
#
# Copyright (c) 2023 Luke Sheneman
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################


from multiprocessing import Pool
import cv2
import math, random
import numpy as np
import sys,os


#
#  !!!!!    IMPORTANT     !!!!!
#
# The number of frames represented by a single frame in our stack.
# Sometimes these centroid stacks are presented to me partially pre-integrated 
# to conserve space.  For example, one frame in the stack could represent 5 
# integrated centroid frames.  Typically this value is 1
SAMPLES_PER_FRAME = 1     

if(len(sys.argv)!=5):
	print("ERROR:  Incorrect number of parameters")
	print("USAGE:  python bootstrap2.py <input_dir> <bootstrap_dir> <bin> <threads>")
	exit(-1)

INPUT_DIR     = sys.argv[1]
BOOTSTRAP_DIR = sys.argv[2]
bin   	      = int(sys.argv[3])
threads       = int(sys.argv[4])

np.set_printoptions(threshold=np.inf)

# The number of times we will subsample our stack
NUM_REPLICATES = 10

SUBSAMPLE_SIZE = int(bin/SAMPLES_PER_FRAME)

print("Input Directory = %s" %INPUT_DIR)
print("Bootstrap Output Directory = %s" %BOOTSTRAP_DIR)
print("Bin = ", bin)
print("SAMPLES_PER_FRAME = ", SAMPLES_PER_FRAME)
print("SUBSAMPLE_SIZE = ", SUBSAMPLE_SIZE)


def chunks(xs, n):
	n = max(1, n)
	return (xs[i:i+n] for i in range(0, len(xs), n))



def bootstrap(chunk):
	print("In bootstrap().  Chunk = %d" %(len(chunk)))
	for f in chunk:
		print("\nReading Cropped Centroid Z-Stack Image..." + f)
		fullpath_in   = os.path.join(INPUT_DIR, f)
		filename_component, file_extension = os.path.splitext(f)

		ret, images = cv2.imreadmulti(fullpath_in, flags=cv2.IMREAD_GRAYSCALE)
		nframes = len(images)
		print("\nImage Read.  Number of time steps = %d" %nframes)	
		for i in range(0,NUM_REPLICATES):
			print("PROCESSING %d of %d BOOTSTRAP REPLICATES" %(i,NUM_REPLICATES))

			bootstrap_subdir = os.path.join(BOOTSTRAP_DIR, str(bin))
			os.makedirs(bootstrap_subdir, exist_ok=True)
			bootstrap_out = os.path.join(bootstrap_subdir, filename_component+"_"+str(i)+file_extension)
	
			integrated_image = np.zeros(images[0].shape, np.uint16)

			# Build a new Sampled Z-Stack
			if(SUBSAMPLE_SIZE <= len(images)):
				sample_stack = random.sample(images, SUBSAMPLE_SIZE)
			else:
				sample_stack = images

			print("   Integrating photons...")
			integrated_arr = np.dstack(sample_stack)
			integrated_image = np.sum(integrated_arr,axis=2,dtype=np.uint8)
			maxval = np.amax(integrated_image)
			print("   Max Photon Count in Integrated Image: %d" %maxval)

			print("   Writing bootstrapped integrated image: %s" %bootstrap_out)
			cv2.imwrite(bootstrap_out, integrated_image)




########################################
#
# Main Execution Section
#
#
def main():
	files = os.listdir(INPUT_DIR)
	ch = chunks(files,math.ceil(len(files)/threads))

	with Pool(threads) as p:
		results = p.map(bootstrap, ch)

if __name__ == '__main__':
	main()


print("Done.")
