##################################################################################
#
# centroid.py
#
# Takes as input a single tiff stack that represents the sensing of 
# photons over time. For every frame in the stack, find the centroid of 
# every photon cloud (blob) and find the intensity "center of mass" moment
# to construct a tiff stack that only contains centroids.
#
# Also integrate the centroid stack to produce a single frame integrated image.
#
##################################################################################
#
# Luke Sheneman, 2023
# sheneman@uidaho.edu
#
# Institute for Interdisciplinary Data Studies (IIDS)
# University of Idaho
#
##################################################################################
#
# MIT License
#
# Copyright (c) 2023 Luke Sheneman
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################


import cv2
import math
import numpy as np
import os


np.set_printoptions(threshold=np.inf)

# set up some paths
ROOT_DIR  = "./data"
INPUT_DIR  = os.path.join(ROOT_DIR, "photonclouds")
CENTROIDS_DIR = os.path.join(ROOT_DIR, "centroids")

OUTBITS = 16    # either 8 or 16


print("Input Directory = %s" %INPUT_DIR)
print("Centroids Directory = %s" %CENTROIDS_DIR)
print("Output Type: %d bit TIFF" %OUTBITS)




########################################
#
# Main Execution Section
#
#

print("Finding centroids...")
files = os.listdir(INPUT_DIR)
print("Processing %d files..." %len(files))
for f in files:

	print("\nReading Z-Stack Image..." + f)
	fullpath_in  = os.path.join(INPUT_DIR, f)
	fullpath_centroids = os.path.join(CENTROIDS_DIR, f)
	ret, images = cv2.imreadmulti(fullpath_in, flags=cv2.IMREAD_GRAYSCALE)
	images = list(images)
	print(ret)
	print(type(images))
	print(len(images))

	new_images = images.copy()
	integrated_image = np.zeros(images[0].shape, np.uint16)

	for i in range(len(images)):

		img = images[i]
		foo = img
		new_img = np.zeros(img.shape, np.uint8)

		# find contours for our masks
		contours_img = cv2.normalize(img,None,0,255,cv2.NORM_MINMAX,cv2.CV_8U)
		contours,hierarchy = cv2.findContours(contours_img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
		#print("Num of Contours = %d" %(len(contours)))
		if(len(contours)==0):
			new_images[i] = new_img
			continue

		for c in contours:

			# calculate moments for each contour
			M = cv2.moments(c)

			# calculate x,y coordinate of center
			if(M["m00"] > 0.0):
				cX = int(M["m10"] / M["m00"])
				cY = int(M["m01"] / M["m00"])

				new_img[cY,cX] = 1

		new_images[i] = new_img

	print("Integrating photons...")
	integrated_arr = np.dstack(new_images)
	integrated_image = np.sum(integrated_arr,axis=2,dtype=np.uint16)
	maxval = np.amax(integrated_image)
	print("Max Photon Count in Integrated Image: %d" %maxval)

	if(OUTBITS==8):
		integrated_image = (integrated_image).astype(np.uint8)
	else:
		integrated_image = (integrated_image).astype(np.uint16)

	print("Writing centroid output...")
	cv2.imwritemulti(fullpath_centroids, new_images)

	print("")
	print("Done.")

