##################################################################################
#
# fli.py
#
# Take detected photon clouds represented as a 12-bit raw stack in FLI format 
# and convert to a TIFF stack representing computed centroids of that stack. 
# Output both the centroid stack as well as a single image representing the 
# integrated (summed Z-projection) of the centroid stack.
#
##################################################################################
#
# Luke Sheneman, 2023
# sheneman@uidaho.edu
#
# Institute for Interdisciplinary Data Studies (IIDS)
# University of Idaho
#
##################################################################################
#
# MIT License
#
# Copyright (c) 2023 Luke Sheneman
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

from multiprocessing import Pool
from flifile import FliFile
import cv2
import math, random
import numpy as np
import sys, os


print(sys.argv)
if(len(sys.argv)!=5):
	print("ERROR:  Incorrect number of parameters")
	print("USAGE:  python fli2.py <input_dir> <input_file> <output_centroids> <output_integrated>")
	exit(-1)

INPUT_DIR         = sys.argv[1]
INPUT_FILE 	  = sys.argv[2]
OUTPUT_CENTROIDS  = sys.argv[3]
OUTPUT_INTEGRATED = sys.argv[4]

print("INPUT_DIR:", INPUT_DIR)
print("INPUT_FILE: ", INPUT_FILE)
print("OUTPUT_CENTROIDS: ",  OUTPUT_CENTROIDS)
print("OUTPUT_INTEGRATED: ", OUTPUT_INTEGRATED)

os.makedirs(OUTPUT_CENTROIDS,  exist_ok=True) 
os.makedirs(OUTPUT_INTEGRATED, exist_ok=True) 


def image_centroids(img):
    centroids = np.zeros(img.shape, np.uint8)

    # find contours for our masks
    contours_img = cv2.normalize(img,None,0,255,cv2.NORM_MINMAX,cv2.CV_8U)

    mean, stdev = cv2.meanStdDev(contours_img)
    mean  = mean.item()
    stdev = stdev.item()	

    edges = cv2.Canny(contours_img, mean-(mean*0.25), mean+stdev)
    #cv2.imwrite("img.tif", contours_img)
    #cv2.imwrite("edges.tif", edges)
    contours,hierarchy = cv2.findContours(edges, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    #print("Num of Contours = %d" %(len(contours)))

    for c in contours:

        # calculate moments for each contour
        M = cv2.moments(c)

        # calculate x,y coordinate of center
        if(M["m00"] > 0.0):
            cX = int(M["m10"] / M["m00"])
            cY = int(M["m01"] / M["m00"])

            centroids[cY,cX] = 1

    #cv2.imwrite("centroids.tif", centroids)
    #exit(0)

    return(centroids)





def centroid_list(cloud_array):
    print("In centroid_stack()...")
    
    centroid_list = []

    (height,width,nframes) = cloud_array.shape
    
    for f in range(nframes):
            centroid_list.append(image_centroids(cloud_array[:,:,f]))

    return(centroid_list)
	



def photon_clouds():
    print("In photon_cloud().  INPUT_FILE = ", INPUT_FILE) 
    print("\nReading Photon Cloud FLI stack..." + INPUT_FILE)
    fullpath_in = os.path.join(INPUT_DIR, INPUT_FILE) 
    filename_component, file_extension = os.path.splitext(INPUT_FILE)

    print(fullpath_in)
    myflifile   = FliFile(fullpath_in)
    cloud_array = myflifile.getdata()

    (height,width,nframes) = cloud_array.shape
    print("Read image %s with shape Width=%d, Height=%d, Frames=%d" %(INPUT_FILE,width,height,nframes))

    print("Writing centroid output...")
    clist = centroid_list(cloud_array)
    fullpath_centroids = os.path.join(OUTPUT_CENTROIDS, filename_component+".tif")
    cv2.imwritemulti(fullpath_centroids, clist)

    print("Integrating centroids...")
    integrated_array = np.dstack(clist)
    integrated_image = np.sum(integrated_array,axis=2,dtype=np.uint16)
    maxval = np.amax(integrated_image)
    print("Max Photon Count in Integrated Image: %d" %maxval)
    fullpath_integrated = os.path.join(OUTPUT_INTEGRATED, filename_component+".tif")
    cv2.imwrite(fullpath_integrated, integrated_image)


photon_clouds() 

print("Done.")

