##################################################################################
#
# integrate.py
#
# Parellelized z-projection across a stack of non-normalized 
# centroids (sum across the z-axis)
#
##################################################################################
#
# Luke Sheneman, 2023
# sheneman@uidaho.edu
#
# Institute for Interdisciplinary Data Studies (IIDS)
# University of Idaho
#
##################################################################################
#
# MIT License
#
# Copyright (c) 2023 Luke Sheneman
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################


from multiprocessing import Pool
import cv2
import math, random
import numpy as np
import sys,os

if(len(sys.argv)!=4):
	print("ERROR:  Incorrect number of parameters")
	print("USAGE:  python <input-dir> <output-dir> <workers>")
	exit(-1)

input_dir  = sys.argv[1]
output_dir = sys.argv[2]
workers    = int(sys.argv[3])

np.set_printoptions(threshold=np.inf)

print("Input Directory = %s" %input_dir)
print("Output Directory = %s" %output_dir)
print("Num Workers = %s" %workers)


def chunks(xs, n):
	n = max(1, n)
	return (xs[i:i+n] for i in range(0, len(xs), n))


def integrate(chunk):
	print("In integrate().  Chunk = %d" %(len(chunk)))
	for f in chunk:
		fullpath_in  = os.path.join(input_dir, f)
		fullpath_out = os.path.join(output_dir, f)
		print("\nReading Centroid Z-Stack Image..." + f)
		ret, raw_images = cv2.imreadmulti(fullpath_in, flags=cv2.IMREAD_UNCHANGED)
		nframes = len(raw_images)
		print("\nZ-Stack Read.  Number of time steps = %d" %nframes)	

		print("Integrating photons from centroids...")
		integrated_arr = np.dstack(raw_images)
		integrated_image = np.sum(integrated_arr,axis=2,dtype=np.uint8)

		maxval = np.amax(integrated_image)
		print("Max Photon Count in Integrated Image: %d" %maxval)

		print("Writing integrated output...")
		cv2.imwrite(fullpath_out, integrated_image)



########################################
#
# Main Execution Section
#
#
def main():
	files = os.listdir(input_dir)
	ch = chunks(files,math.ceil(len(files)/workers))

	with Pool(workers) as p:
		results = p.map(integrate, ch)

if __name__ == '__main__':
	main()

print("Done.")

