##################################################################################
#
# metrics.py
#
# Compute similarity/distance metrics between all images in two specified 
# folders or between two files
#
##################################################################################
#
# Luke Sheneman, 2023
# sheneman@uidaho.edu
#
# Institute for Interdisciplinary Data Studies (IIDS)
# University of Idaho
#
##################################################################################
#
# MIT License
#
# Copyright (c) 2023 Luke Sheneman
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
#################################################################################

import os, sys
import cv2
import numpy as np
from skimage.metrics import structural_similarity as ssim
from skimage.metrics import mean_squared_error as mse
from skimage.metrics import normalized_root_mse as nrmse
from skimage.metrics import peak_signal_noise_ratio as psnr
from scipy.stats import pearsonr  # pearson correlation coefficient

DATADIR  = "data"
RUNSDIR  = "runs"
#DIRBASE  = "8depth_16filter_1000train_"
DIRBASE  = "4depth_4filter_1000train_"
FRAMES   = [1,2,3,4,5,10,15,20,25,30,50,75,100,125,150,200,250,500,750,1000,2000,3000,4000]
#FRAMES   = [1,2]
#DIRS     = ["0004_exp_roots", "0005_exp_yeast", "0006_yeast2", "0007_yeast3", "0008_root_examples"]
DIRS     = ["0004_exp_roots"]

for dir in DIRS:
	mdict = {}

	for frame in FRAMES:

		total_ssim    = 0
		total_mse     = 0
		total_nrmse   = 0
		total_psnr    = 0
		total_pearson = 0

		pred_input = os.path.join(RUNSDIR, dir, DIRBASE+str(frame)+"frames_shallow", "test32")
		true_input = os.path.join(DATADIR, dir, "purelet8")


		files = os.listdir(pred_input)
		for f in files:

			if not f in mdict:
				mdict[f]={}

			fn = str(f)
			pred_fp = os.path.join(pred_input, fn)
			true_fp = os.path.join(true_input, fn)

			print(pred_fp)

			pred_img = cv2.imread(pred_fp,flags=cv2.IMREAD_UNCHANGED)
			true_img = cv2.imread(true_fp,flags=cv2.IMREAD_UNCHANGED)

			pred_img = cv2.normalize(pred_img,None,0,255,cv2.NORM_MINMAX,cv2.CV_8U)
			true_img = cv2.normalize(true_img,None,0,255,cv2.NORM_MINMAX,cv2.CV_8U)

			mdict[f][str(frame)+".ssim"]       = ssim(true_img, pred_img)
			mdict[f][str(frame)+".mse"]    	= mse(true_img, pred_img)
			mdict[f][str(frame)+".nrmse"]  	= nrmse(true_img, pred_img, normalization='euclidean')
			mdict[f][str(frame)+".(1-nrmse)"]  = 1.0-nrmse(true_img, pred_img, normalization='euclidean')
			mdict[f][str(frame)+".psnr"] 	= psnr(true_img, pred_img)
			mdict[f][str(frame)+".pearson"],x  = pearsonr(true_img.flatten(), pred_img.flatten())
			mdict[f][str(frame)+".pearson2"] = np.corrcoef(true_img, pred_img)

			a,x = pearsonr(true_img.flatten(), pred_img.flatten())
			b = np.corrcoef(true_img.flatten(), pred_img.flatten())

			print(a, b[0,1], b[1,0])


	fn = dir + "_shallow2.csv"
	outfile = open(fn, "w")
	outfile.write("filename")
	k = list(mdict.keys())[0]
	for v in mdict[k]:
		outfile.write(","+v)
	outfile.write("\n")

	for fkey in mdict:
		outfile.write(fkey)
		for v in mdict[fkey]:
			outfile.write(","+str(mdict[fkey][v]))
		outfile.write("\n")
	outfile.close()

	
